import numpy as np
import scipy as sp

from pygradflow.iterate import Iterate
from pygradflow.linear_solver import LinearSolverError
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.step_solver_error import StepSolverError

from .scaled_step_solver import ScaledStepSolver


class SymmetricStepSolver(ScaledStepSolver):
    def __init__(
        self,
        problem: Problem,
        params: Params,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
    ) -> None:
        super().__init__(problem, params, orig_iterate, dt, rho)

        assert dt > 0.0
        assert rho > 0.0

    def compute_hess_jac(self) -> None:
        inactive_indices = np.where(np.logical_not(self.active_set))[0]

        lamb = 1.0 / self.dt

        n = self.n

        hess = self._hess + sp.sparse.diags(
            [lamb], shape=(n, n), dtype=self.params.dtype
        )

        hess_rows = hess.tocsr()[inactive_indices, :]
        self.hess_rows = hess_rows.tocsc()

    def update_derivs(self, iterate: Iterate) -> None:
        super().update_derivs(iterate)
        self._jac = self.jac.tocsc()

    def reset_deriv(self) -> None:
        super().reset_deriv()
        self.hess_rows = None

    def _compute_deriv(self, active_set: np.ndarray) -> sp.sparse.spmatrix:
        inactive_indices = np.where(np.logical_not(self.active_set))[0]
        lamb = 1.0 / self.dt
        rho = self.rho

        m = self.m

        inactive_jac = self.jac[:, inactive_indices]
        inactive_hess = self.hess_rows[:, inactive_indices]

        lower_mat = sp.sparse.diags(
            [-lamb / (1.0 + lamb * rho)], shape=(m, m), dtype=self.params.dtype
        )

        deriv = sp.sparse.bmat(
            [
                [inactive_hess, inactive_jac.T],
                [inactive_jac, lower_mat],
            ],
            format="csc",
        )

        assert deriv.dtype == self.params.dtype

        # May not be the case if Hessian itself is not
        # numerically symmetric
        # assert np.allclose((deriv - deriv.T).data, 0.0)

        return deriv

    def compute_rhs(
        self,
        active_indices: np.ndarray,
        b0: np.ndarray,
        b1: np.ndarray,
        b2t: np.ndarray,
    ) -> np.ndarray:
        active_hess = self.hess_rows[:, active_indices]
        active_jac = self.jac[:, active_indices]

        b1c = b1 - (active_hess @ b0)
        b2c = b2t - (active_jac @ b0)

        rhs = np.concatenate((b1c, b2c))

        return rhs

    def solve_scaled(self, b0, b1, b2t):
        params = self.params

        n = self.n
        m = self.m

        active_indices = np.where(self.active_set)[0]
        inactive_indices = np.where(np.logical_not(self.active_set))[0]

        inactive_set_size = inactive_indices.size

        self.compute_hess_jac()

        rhs = self.compute_rhs(active_indices, b0, b1, b2t)

        assert rhs.shape == (inactive_set_size + m,)

        s = self._solve_active_set(self.active_set, rhs)

        inactive_dx = s[:inactive_set_size]
        dy = s[inactive_set_size:]

        dx = np.zeros((n,), dtype=self.params.dtype)

        dx[inactive_indices] = inactive_dx
        dx[active_indices] = b0

        rcond = None
        if params.report_rcond:
            rcond = self.estimate_rcond(self.deriv, self.solver)

        return (dx, dy, rcond)

    def linear_solver(self, mat: sp.sparse.spmatrix):
        from pygradflow.linear_solver import linear_solver

        solver_type = self.params.linear_solver_type
        return linear_solver(mat, solver_type, symmetric=True)

    def _solve_deriv(
        self, active_set: np.ndarray, deriv: sp.sparse.spmatrix, rhs: np.ndarray
    ) -> np.ndarray:
        problem = self.problem
        params = self.params

        try:
            if self.solver is None:
                self.solver = self.linear_solver(self.deriv)
            sol = self.solver.solve(rhs)

            if params.inertia_correction:
                num_neg_eigvals = self.solver.num_neg_eigvals()

                if num_neg_eigvals is None:
                    raise Exception("Inertia correction requested but not available")

                if num_neg_eigvals != problem.num_cons:
                    raise LinearSolverError("Invalid matrix inertia")

        except LinearSolverError as e:
            raise StepSolverError from e

        return sol

    def _solve_active_set(self, active_set: np.ndarray, rhs: np.ndarray) -> np.ndarray:
        if self._deriv is None:
            self._deriv = self._compute_deriv(self.active_set)

        return self._solve_deriv(self.active_set, self.deriv, rhs)



import copy

import numpy as np
import scipy as sp

from pygradflow.iterate import Iterate
from pygradflow.linear_solver import LinearSolverError
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.step.step_solver_error import StepSolverError

from .scaled_step_solver import ScaledStepSolver


class AsymmetricStepSolver(ScaledStepSolver):
    def __init__(
        self,
        problem: Problem,
        params: Params,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
    ) -> None:
        super().__init__(problem, params, orig_iterate, dt, rho)

        assert dt > 0.0
        assert rho > 0.0

    def update_derivs(self, iterate: Iterate) -> None:
        self._jac = copy.copy(iterate.aug_lag_deriv_xy())
        self._hess = copy.copy(iterate.aug_lag_deriv_xx(rho=0.0))
        self.reset_deriv()

    def update_active_set(self, active_set: np.ndarray) -> None:
        self._active_set = copy.copy(active_set)
        self.reset_deriv()

    def overwrite_active_rows(self, matrix):
        m = self.m
        n = self.n

        active_set = self.active_set

        assert matrix.shape == (n + m, n + m)

        data = matrix.data
        cols = matrix.indices
        col_index = matrix.indptr

        for j in range(n):
            active = active_set[j]

            if not active:
                continue

            col_start = col_index[j]
            col_end = col_index[j + 1]

            curr_data = data[col_start:col_end]
            curr_cols = cols[col_start:col_end]

            assert curr_data.shape == curr_cols.shape

            (col_nnz,) = curr_data.shape

            k = np.searchsorted(curr_cols, j)

            curr_data[:] = 0.0
            curr_data[k] = 1.0
            curr_cols[k] = j

            assert (curr_cols[:-1] <= curr_cols[1:]).all()
            assert (0 <= curr_cols).all()
            assert (curr_cols < n + m).all()

    def compute_deriv(self, active_set: np.ndarray) -> sp.sparse.spmatrix:
        lamb = 1.0 / self.dt
        rho = self.rho

        m = self.m
        n = self.n

        jac = self.jac
        hess = self.hess

        hess += sp.sparse.diags([lamb], shape=(n, n), dtype=self.params.dtype)

        lower_mat = sp.sparse.diags(
            [-lamb / (1.0 + lamb * rho)], shape=(m, m), dtype=self.params.dtype
        )

        deriv = sp.sparse.bmat(
            [
                [hess, jac.T],
                [jac, lower_mat],
            ],
            format="csr",
        )

        self.overwrite_active_rows(deriv)

        assert deriv.dtype == self.params.dtype

        return deriv

    def compute_rhs(self, b0, b1, b2t):
        m = self.m
        n = self.n

        params = self.params
        active_set = self.active_set

        rhs = np.empty((n + m,), dtype=params.dtype)

        if m > 0:
            rhs[-m:] = b2t

        var_rhs = rhs[:n]

        var_rhs[active_set] = b0
        var_rhs[np.logical_not(active_set)] = b1

        return rhs

    def initial_sol(self, b0, b1, b2t):
        m = self.m
        n = self.n

        params = self.params
        active_set = self.active_set

        def initial_sol():
            sol = np.zeros((n + m,), dtype=params.dtype)
            var_sol = sol[:n]
            var_sol[active_set] = b0
            return sol

        return initial_sol

    def solve_scaled(self, b0, b1, b2t):
        if self._deriv is None:
            self._deriv = self.compute_deriv(self.active_set)

        if self.solver is None:
            self.solver = self.linear_solver(self.deriv)

        params = self.params

        rhs = self.compute_rhs(b0, b1, b2t)

        initial_sol = self.initial_sol(b0, b1, b2t)

        try:
            sol = self.solver.solve(rhs, initial_sol=initial_sol)
        except LinearSolverError as e:
            raise StepSolverError from e

        m = self.m
        n = self.n

        assert sol.shape == (m + n,)

        var_sol = sol[:n]
        cons_sol = sol[n:]

        rcond = None
        if params.report_rcond:
            try:
                rcond = self.estimate_rcond(self.deriv, self.solver)
            except LinearSolverError:
                pass

        return (var_sol, cons_sol, rcond)



from abc import ABC, abstractmethod
from typing import Optional

import scipy as sp
from numpy import ndarray


class LinearSolverError(Exception):
    """
    Error signaling that the linear solver failed, e.g. because the
    matrix is (near) singular. The solver attempts to recover by
    reducing the step size
    """

    pass


class LinearSolver(ABC):

    def __init__(self, matrix: sp.sparse.spmatrix, symmetric=False):
        self.symmetric = symmetric

    @abstractmethod
    def solve(self, rhs: ndarray, trans: bool = False, initial_sol=False) -> ndarray:
        raise NotImplementedError()

    def num_neg_eigvals(self) -> Optional[int]:
        return None

    def rcond(self) -> Optional[float]:
        return None



from enum import Enum, auto


class SolverStatus(Enum):
    Optimal = auto()
    """
    The algorithm has converged to a solution satisfying
    the optimality conditions according to given tolerances
    """

    IterationLimit = auto()
    """
    Reached the iteration limit precribed by the algorithmic
    parameters
    """

    TimeLimit = auto()
    """
    Reached the time limit precribed by the algorithmic
    parameters
    """

    Unbounded = auto()
    """
    Problem appearst unbounded (found feasible point with extremely
    small objective value)
    """

    LocallyInfeasible = auto()
    """
    Local infeasibility detected (found infeasible point being
    a local minimum with respect to constraint violation)
    """

    @staticmethod
    def short_name(status):
        return {
            SolverStatus.Optimal: "optimal",
            SolverStatus.IterationLimit: "iteration_limit",
            SolverStatus.TimeLimit: "time_limit",
            SolverStatus.Unbounded: "unbounded",
            SolverStatus.LocallyInfeasible: "infeasible",
        }[status]

    @staticmethod
    def description(status):
        return {
            SolverStatus.Optimal: "Converged to first-order optimal solution",
            SolverStatus.IterationLimit: "Reached iteration limit",
            SolverStatus.TimeLimit: "Reached time limit",
            SolverStatus.Unbounded: "Problem appears unbounded",
            SolverStatus.LocallyInfeasible: "Local infeasibility detected",
        }[status]

    @staticmethod
    def success(status):
        """
        Returns
        -------
        bool
            Whether the status indicates a successful solve
        """
        return status == SolverStatus.Optimal



import numpy as np
import scipy as sp

from pygradflow.integration.flow import Flow
from pygradflow.integration.problem_switches import ProblemSwitches
from pygradflow.iterate import Iterate
from pygradflow.util import keep_rows


class RestrictedFlow:
    def __init__(self, flow, filter):
        self.flow = flow
        self.problem = flow.problem
        self.params = flow.params
        self.eval = flow.eval
        problem = flow.problem
        num_vars = problem.num_vars

        assert filter.shape == (num_vars,)
        assert filter.dtype == bool

        self.filter = filter
        self.switches = ProblemSwitches(self)

    def check_point(self, z, rho):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub

        (x, y) = self.flow.split_states(z)
        at_lb = Flow.isclose(x, lb)
        at_ub = Flow.isclose(x, ub)

        rhs = self.rhs(z, rho)
        (rhs_x, _) = self.flow.split_states(rhs)

        num_vars = problem.num_vars
        filter = self.filter

        for j in range(num_vars):
            if at_lb[j]:
                if rhs_x[j] < 0.0 and not (Flow.isclose(rhs_x[j], 0.0)):
                    assert not filter[j]
                elif rhs_x[j] > 0.0 and not (Flow.isclose(rhs_x[j], 0.0)):
                    assert filter[j]
            elif at_ub[j]:
                if rhs_x[j] > 0.0 and not (Flow.isclose(rhs_x[j], 0.0)):
                    assert not filter[j]
                elif rhs_x[j] < 0.0 and not (Flow.isclose(rhs_x[j], 0.0)):
                    assert filter[j]
            else:
                assert filter[j]

    def create_event_triggers(self, z, rho):
        return self.switches.create_event_triggers(self.filter, z, rho)

    def rhs_deriv_x(self, z, rho, c=None):
        problem = self.problem
        params = self.params
        eval = self.eval

        (x, y) = self.flow.split_states(z)

        iterate = Iterate(problem, params, x, y)
        hess = iterate.aug_lag_deriv_xx(rho)
        jac = iterate.aug_lag_deriv_xy()

        if c is None:
            c = eval.cons(x)

        aug_lag_deriv_x = self.flow.aug_lag_deriv_x(z, rho, c=c)
        # aug_lag_deriv_x *= self.filter
        aug_lag_deriv_y = c

        dt = hess @ aug_lag_deriv_x - jac.T @ aug_lag_deriv_y
        return dt

    def rhs(self, z, rho, c=None):
        (x, y) = self.flow.split_states(z)
        eval = self.eval

        if c is None:
            c = eval.cons(x)

        dx = self.flow.neg_aug_lag_deriv_x(z, rho, c)
        dx *= self.filter
        dy = c
        return np.concatenate((dx, dy))

    def rhs_func(self, rho):
        assert rho > 0

        def rhs(_, z):
            return self.rhs(z, rho)

        return rhs

    def rhs_jac(self, z, rho):
        problem = self.problem
        params = self.params

        (x, y) = self.flow.split_states(z)

        iterate = Iterate(problem, params, x, y)
        hess = iterate.aug_lag_deriv_xx(rho)
        jac = iterate.aug_lag_deriv_xy()

        filtered_hess = keep_rows(hess, self.filter)
        filtered_jac = keep_rows(jac.T, self.filter)

        return sp.sparse.bmat(
            [[-filtered_hess, -filtered_jac], [jac, None]], format="csr"
        )

    def rhs_jac_func(self, rho):
        assert rho > 0

        def rhs_jac(_, z):
            return self.rhs_jac(z, rho)

        return rhs_jac

    def residuum(self, z):
        return np.linalg.norm(self.rhs(z, rho=0.0))



import numpy as np

from pygradflow.iterate import Iterate

EPS = np.finfo(float).eps
xtol = 4 * EPS
atol = 4 * EPS


def lazy_func(func):
    values = [None]

    def wrapped(*args, **kwds):
        if values[0] is None:
            values[0] = func(*args, **kwds)
        return values[0]

    return wrapped


def is_pos(x):
    return x > 0.0 and not Flow.isclose(x, 0.0)


def is_neg(x):
    return x < 0.0 and not Flow.isclose(x, 0.0)


def func_pos(func, deriv, j):
    if is_pos(func()[j]):
        return True
    elif Flow.isclose(func()[j], 0.0):
        if is_pos(deriv()[j]):
            return True
    return False


def func_neg(func, deriv, j):
    if is_neg(func()[j]):
        return True
    elif Flow.isclose(func()[j], 0.0):
        if is_neg(deriv()[j]):
            return True
    return False


class Flow:
    def __init__(self, problem, params, eval):
        self.problem = problem
        self.params = params
        self.eval = eval

    @staticmethod
    def isclose(x, y):
        return np.isclose(x, y, rtol=xtol, atol=atol)

    @property
    def num_vars(self):
        return self.problem.num_vars

    @property
    def num_cons(self):
        return self.problem.num_cons

    def is_boxed(self, x):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub
        return (lb <= x).all() and (x <= ub).all()

    def is_approx_boxed(self, x):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub

        is_boxed_lb = np.logical_or(lb <= x, Flow.isclose(lb, x))
        is_boxed_ub = np.logical_or(x <= ub, Flow.isclose(ub, x))

        return np.logical_and(is_boxed_lb, is_boxed_ub).all()

    def rhs(self, z, rho, c=None):
        eval = self.eval
        (x, y) = self.split_states(z)

        if c is None:
            c = eval.cons(x)

        dx = self.neg_aug_lag_deriv_x(z, rho, c)
        dy = c
        return np.concatenate((dx, dy))

    def rhs_func(self, rho):
        def rhs(t, z):
            return self.rhs(z, rho)

        return rhs

    def aug_lag_deriv_x(self, z, rho, c=None):
        eval = self.eval
        (x, y) = self.split_states(z)

        if c is None:
            c = eval.cons(x)

        lhs = rho * c + y
        aug_lag_deriv_x = eval.obj_grad(x) + eval.cons_jac(x).T.dot(lhs)
        return aug_lag_deriv_x

    def aug_lag_deriv_y(self, z, rho, c=None):
        eval = self.eval
        (x, y) = self.split_states(z)

        if c is None:
            c = eval.cons(x)

        return c

    def neg_aug_lag_deriv_x(self, z, rho, c=None):
        return -self.aug_lag_deriv_x(z, rho, c)

    def obj(self, z):
        problem = self.problem
        (x, y) = self.split_states(z)
        return problem.obj(x)

    def neg_aug_lag_deriv_xx(self, z, rho, c=None):
        eval = self.eval
        problem = self.problem
        params = self.params

        (x, y) = self.split_states(z)

        iterate = Iterate(problem, params, x, y)
        hess = iterate.aug_lag_deriv_xx(rho)
        jac = iterate.aug_lag_deriv_xy()

        if c is None:
            c = eval.cons(x)

        aug_lag_deriv_x = self.aug_lag_deriv_x(z, rho, c=c)
        aug_lag_deriv_y = c

        dt = hess @ aug_lag_deriv_x - jac.T @ aug_lag_deriv_y
        return dt

    def rhs_deriv_x(self, z, rho):
        return self.neg_aug_lag_deriv_xx(z, rho)

    def split_states(self, z):
        problem = self.problem
        num_vars = problem.num_vars
        num_cons = problem.num_cons
        assert z.shape == (num_vars + num_cons,)

        x = z[:num_vars]
        y = z[num_vars:]
        return x, y



import copy
from typing import Optional, Tuple, cast

import numpy as np
import scipy as sp

from pygradflow.implicit_func import ScaledImplicitFunc
from pygradflow.iterate import Iterate
from pygradflow.params import Params
from pygradflow.problem import Problem

from .step_solver import StepResult, StepSolver


class ScaledStepSolver(StepSolver):
    def __init__(
        self,
        problem: Problem,
        params: Params,
        orig_iterate: Iterate,
        dt: float,
        rho: float,
    ) -> None:
        super().__init__(problem, params)

        self.orig_iterate = orig_iterate
        self._func = ScaledImplicitFunc(problem, orig_iterate, dt)

        self._deriv: Optional[sp.sparse.spmatrix] = None

        self.dt = dt
        self.rho = rho

    @property
    def func(self) -> ScaledImplicitFunc:
        return self._func

    def initial_rhs(
        self, iterate: Iterate
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        n = self.n
        m = self.m

        rhs = self.func.value_at(iterate, self.rho, self.active_set)

        assert rhs.shape == (n + m,)

        rx = rhs[:n]
        ry = rhs[n:]

        active_indices = np.where(self.active_set)[0]
        inactive_indices = np.where(np.logical_not(self.active_set))[0]

        dt = self.dt

        b0 = dt * rx[active_indices]
        b1 = rx[inactive_indices]
        b2 = ry

        return (b0, b1, b2)

    def solve_scaled(
        self, b0, b1, b2t
    ) -> Tuple[np.ndarray, np.ndarray, Optional[float]]:
        raise NotImplementedError()

    def reset_deriv(self) -> None:
        self._deriv = None
        self.solver = None

    @property
    def deriv(self) -> sp.sparse.spmatrix:
        assert self._deriv is not None
        return cast(sp.sparse.spmatrix, self._deriv)

    def update_derivs(self, iterate: Iterate) -> None:
        self._jac = copy.copy(iterate.aug_lag_deriv_xy())
        self._hess = copy.copy(iterate.aug_lag_deriv_xx(rho=0.0))
        self.reset_deriv()

    def update_active_set(self, active_set: np.ndarray) -> None:
        self._active_set = copy.copy(active_set)
        self.reset_deriv()

    def solve(self, iterate: Iterate) -> StepResult:
        (b0, b1, b2) = self.initial_rhs(iterate)

        n = self.n
        m = self.m

        rho = self.rho
        lamb = 1.0 / self.dt
        fact = 1.0 / (1.0 + lamb * rho)

        assert fact > 0.0

        b2t = fact * b2

        (sx, sy, rcond) = self.solve_scaled(b0, b1, b2t)

        assert sx.shape == (n,)
        assert sy.shape == (m,)

        dx = sx
        dy = fact * (sy - rho * b2)

        return StepResult(iterate, dx, dy, self.active_set, rcond)



import math
from dataclasses import dataclass

from pygradflow.params import Params


@dataclass
class ControllerSettings:
    K_P: float = 0.0
    K_I: float = 0.0

    lamb_init: float = 0.0
    lamb_red: float = 0.0

    def __post_init__(self) -> None:
        assert self.K_P >= 0.0
        assert self.K_I >= 0.0

    @staticmethod
    def from_params(params: Params) -> "ControllerSettings":
        return ControllerSettings(
            K_P=params.K_P,
            K_I=params.K_I,
            lamb_init=params.lamb_init,
            lamb_red=params.lamb_red,
        )


class Controller:
    """
    PI controller with given settings and reference value.
    """

    def __init__(self, settings: ControllerSettings, ref: float) -> None:
        self.settings = settings
        self.ref = ref

        self.value = settings.lamb_init
        self.error_sum = 0.0

    def reset(self) -> None:
        self.error_sum = 0.0

    def update(self, val: float) -> float:
        error = self.ref - val
        self.error_sum += error

        update_term = self.settings.K_P * error + self.settings.K_I * self.error_sum

        self.value = update_term
        return self.value


class LogController:
    """
    PI controller working on log scale.
    """

    def __init__(self, settings: ControllerSettings, ref: float) -> None:
        self.settings = settings
        assert ref > 0.0

        self.controller = Controller(settings, math.log(ref))

        self.ref = ref
        self.error_sum = 0.0

    @property
    def value(self) -> float:
        return math.exp(self.controller.value)

    def update(self, val: float) -> float:
        assert val > 0.0

        self.controller.update(math.log(val))

        return self.value



import numpy as np

from pygradflow.implicit_func import ImplicitFunc
from pygradflow.log import logger
from pygradflow.step.newton_control import NewtonController
from pygradflow.step.step_control import StepControlResult
from pygradflow.step.step_solver_error import StepSolverError


class ExactController(NewtonController):
    def __init__(self, problem, params, max_num_it=10, rate_bound=0.5):
        super().__init__(problem, params)
        self.max_num_it = max_num_it
        self.rate_bound = rate_bound

    def step(self, iterate, rho, dt, display, timer):
        assert dt > 0.0
        lamb = 1.0 / dt

        func = ImplicitFunc(self.problem, iterate, dt)

        def func_val(iterate):
            return np.linalg.norm(func.value_at(iterate, rho))

        curr_func_val = func_val(iterate)

        next_steps = self.newton_steps(iterate, rho, dt)

        rcond = None
        active_set = None

        for i in range(self.max_num_it):
            next_step = next(next_steps)
            next_iterate = next_step.iterate
            active_set = next_step.active_set
            rcond = next_step.rcond

            if timer.reached_time_limit():
                raise StepSolverError("Time limit reached")

            self.display_step(i, next_step)

            next_func_val = func_val(next_iterate)
            logger.debug(f"Func val: {next_func_val}")

            if next_func_val <= self.params.newton_tol:
                logger.debug("Newton method converged in %d iterations", i + 1)
                return StepControlResult(
                    next_iterate, 0.5 * lamb, active_set, rcond, True
                )

            rate_est = next_func_val / curr_func_val

            if rate_est > self.rate_bound:
                logger.debug(
                    "Newton convergence rate (%f) exceeded bound (%f)",
                    rate_est,
                    self.rate_bound,
                )
                break

            curr_func_val = next_func_val

        logger.debug("Newton method did not converge in %d iterations", self.max_num_it)

        return StepControlResult(next_iterate, 2.0 * lamb, active_set, rcond, False)



from enum import Enum, auto

import numpy as np

from pygradflow.integration.flow import lazy_func


class TriggerType(Enum):
    LB = auto()
    UB = auto()
    PENALTY = auto()
    GRAD_FIXED = auto()
    CONVERGED = auto()
    UNBOUNDED = auto()


class SwitchTrigger:
    def __init__(self, time, state, trigger):
        self.time = time
        self.state = state
        self.trigger = trigger

    @property
    def type(self):
        return self.trigger.type

    @property
    def index(self):
        return self.trigger.index


class ProblemSwitches:
    def __init__(self, restricted_flow):
        self.restricted_flow = restricted_flow
        self.flow = restricted_flow.flow
        self.problem = restricted_flow.problem
        self.params = restricted_flow.params

    def lb_event(self, j):
        lb = self.problem.var_lb

        def at_lb(_, z):
            (x, y) = self.flow.split_states(z)
            return x[j] - lb[j]

        at_lb.type = TriggerType.LB
        at_lb.index = j
        at_lb.direction = -1.0

        return at_lb

    def ub_event(self, j):
        ub = self.problem.var_ub

        def at_ub(_, z):
            (x, y) = self.flow.split_states(z)
            return x[j] - ub[j]

        at_ub.type = TriggerType.UB
        at_ub.index = j
        at_ub.direction = 1.0

        return at_ub

    def penalty_event(self, rho):
        def penalty_check(_, z):
            rhs = self.restricted_flow.rhs(z, rho)
            rhs_x, rhs_y = self.flow.split_states(rhs)
            aug_dx = self.flow.aug_lag_deriv_x(z, rho)
            aug_dy = self.flow.aug_lag_deriv_y(z, rho)

            return np.dot(rhs_x, aug_dx) + np.dot(rhs_y, aug_dy)

        penalty_check.type = TriggerType.PENALTY
        penalty_check.direction = 1.0

        return penalty_check

    def grad_fixed_event(self, j, rho, at_lb):
        def grad(_, z):
            return self.flow.neg_aug_lag_deriv_x(z, rho)[j]

        grad.type = TriggerType.GRAD_FIXED
        grad.index = j

        if at_lb:
            grad.direction = 1.0
        else:
            grad.direction = -1.0

        return grad

    def converged_event(self):
        params = self.params

        def converged_event(_, z):
            return self.restricted_flow.residuum(z) - params.opt_tol

        converged_event.type = TriggerType.CONVERGED

        return converged_event

    def unbounded_event(self):
        params = self.params

        def unbounded_event(_, z):
            return self.flow.obj(z) - params.obj_lower_limit

        unbounded_event.type = TriggerType.UNBOUNDED

        return unbounded_event

    def create_event_triggers(self, filter, z_curr, rho):
        lb = self.problem.var_lb
        ub = self.problem.var_ub

        problem = self.problem
        num_vars = problem.num_vars

        assert filter.shape == (num_vars,)
        assert filter.dtype == bool

        @lazy_func
        def rhs():
            return self.flow.rhs(z_curr, rho)

        @lazy_func
        def rhs_deriv():
            return self.flow.rhs_deriv_x(z_curr, rho)

        (x, y) = self.flow.split_states(z_curr)

        lb_events = []
        ub_events = []
        grad_fixed_events = []

        at_lb = self.flow.isclose(x, lb)
        at_ub = self.flow.isclose(x, ub)

        for j in range(num_vars):
            # Variable is free
            if filter[j]:

                lb_event = self.lb_event(j)
                ub_event = self.ub_event(j)

                if np.isfinite(lb[j]):
                    lb_events.append(lb_event)

                if np.isfinite(ub[j]):
                    ub_events.append(ub_event)

            # Variable is pinned, watch for sign changes
            if not filter[j]:
                # Equal bounds, flow *must* stay at zero
                if at_lb[j] and at_ub[j]:
                    continue

                assert at_lb[j] or at_ub[j]

                event = self.grad_fixed_event(j, rho, at_lb[j])
                grad_fixed_events.append(event)

        events = [
            *lb_events,
            *ub_events,
            *grad_fixed_events,
            self.converged_event(),
            self.unbounded_event(),
            self.penalty_event(rho),
        ]

        for event in events:
            event.terminal = True

        return events



from typing import Optional

import numpy as np
import scipy as sp

from pygradflow.params import Params, ScalingType
from pygradflow.problem import Problem
from pygradflow.util import sparse_zero


def scale_symmetric(A):
    (n, _) = A.shape

    A = A.tocoo()
    a_rows = A.row
    a_cols = A.col
    a_data = np.abs(A.data)

    max_it = 100

    D = np.zeros((n,), dtype=int)

    for i in range(max_it):
        R = np.zeros((n,), dtype=int)

        for k in range(len(a_data)):
            R[a_cols[k]] += a_data[k]

        R[R < 1e-10] = 1.0
        R = np.sqrt(R)

        Rsca = 1 - np.frexp(R)[1]

        if (Rsca == 0).all():
            break

        for k in range(len(a_data)):
            a_data[k] = np.ldexp(a_data[k], Rsca[a_rows[k]] + Rsca[a_cols[k]])

        D += Rsca
    else:
        raise Exception("Equilibration failed to converge")

    return D


class Scaling:
    def __init__(self, var_weights, cons_weights, obj_weight=0):
        self.var_weights = var_weights
        self.cons_weights = cons_weights

        assert var_weights.ndim == 1
        assert var_weights.dtype in [np.int64, np.int32, np.int16, np.int8]

        assert cons_weights.ndim == 1
        assert cons_weights.dtype in [np.int64, np.int32, np.int16, np.int8]

        self.obj_weight = obj_weight

    @staticmethod
    def zero(num_vars, num_cons):
        return Scaling(
            np.zeros((num_vars,), dtype=int), np.zeros((num_cons,), dtype=int)
        )

    @staticmethod
    def from_nominal_values(var_values, cons_values, obj_value=1.0):
        var_weights = Scaling.weights_from_nominal_values(var_values)
        cons_weights = Scaling.weights_from_nominal_values(cons_values)
        obj_weight = Scaling.weights_from_nominal_values(obj_value)

        return Scaling(var_weights, cons_weights, obj_weight)

    @staticmethod
    def weights_from_nominal_values(values):
        return 1 - np.frexp(values)[1]

    @staticmethod
    def from_grad_jac(obj_grad, cons_jac):
        grad_weights = Scaling.weights_from_nominal_values(np.abs(obj_grad))
        var_weights = -grad_weights

        if cons_jac is None:
            cons_weights = np.zeros((0,), dtype=int)
            return Scaling(var_weights, cons_weights)

        (num_cons, num_vars) = cons_jac.shape
        assert obj_grad.shape == (num_vars,)

        jac = cons_jac.tocoo()

        rows = jac.row
        cols = jac.col
        data = np.abs(jac.data)

        prescaled_data = np.ldexp(data, -var_weights[cols])
        max_values = np.zeros((num_cons,), dtype=int)

        for i, row in enumerate(rows):
            max_values[row] = max(max_values[row], prescaled_data[i])

        cons_weights = Scaling.weights_from_nominal_values(max_values)

        return Scaling(var_weights, cons_weights)

    @staticmethod
    def from_equilibrated_kkt(lag_hess, cons_jac):
        (m, n) = cons_jac.shape
        assert lag_hess.shape == (n, n)

        kkt_mat = sp.sparse.bmat([[lag_hess, cons_jac.T], [cons_jac, None]])

        weights = scale_symmetric(kkt_mat)

        var_weights = -weights[:n]
        cons_weights = weights[n:]

        return Scaling(var_weights, cons_weights)

    @property
    def num_vars(self):
        return len(self.var_weights)

    @property
    def num_cons(self):
        return len(self.cons_weights)

    def scale_primal(self, x):
        return np.ldexp(x, self.var_weights)

    def unscale_primal(self, x):
        return np.ldexp(x, -self.var_weights)

    def _dual_weights(self):
        return self.cons_weights - self.obj_weight

    def _bound_weights(self):
        return self.var_weights - self.obj_weight

    def scale_dual(self, y):
        return np.ldexp(y, -self._dual_weights())

    def unscale_dual(self, y):
        return np.ldexp(y, self._dual_weights())

    def scale_bounds_dual(self, y):
        return np.ldexp(y, -self._bound_weights())

    def unscale_bounds_dual(self, y):
        return np.ldexp(y, self._bound_weights())


class ScaledProblem(Problem):
    def __init__(self, problem: Problem, scaling: Scaling):
        self.problem = problem
        self.scaling = scaling

        var_lb = np.ldexp(problem.var_lb, scaling.var_weights)
        var_ub = np.ldexp(problem.var_ub, scaling.var_weights)

        cons_lb = np.ldexp(problem.cons_lb, scaling.cons_weights)
        cons_ub = np.ldexp(problem.cons_ub, scaling.cons_weights)

        super().__init__(var_lb, var_ub, cons_lb=cons_lb, cons_ub=cons_ub)

    def _orig_x(self, scaled_x):
        var_weights = self.scaling.var_weights
        return np.ldexp(scaled_x, -var_weights)

    def obj(self, x):
        x_orig = self._orig_x(x)
        obj_orig = self.problem.obj(x_orig)
        return np.ldexp(obj_orig, self.scaling.obj_weight)

    def obj_grad(self, x):
        var_weights = self.scaling.var_weights
        x_orig = self._orig_x(x)
        grad_orig = self.problem.obj_grad(x_orig)
        grad = np.ldexp(grad_orig, -var_weights)

        return np.ldexp(grad, self.scaling.obj_weight)

    def cons(self, x):
        cons_weights = self.scaling.cons_weights
        x_orig = self._orig_x(x)

        cons_orig = self.problem.cons(x_orig)

        return np.ldexp(cons_orig, cons_weights)

    def cons_jac(self, x):
        var_weights = self.scaling.var_weights
        cons_weights = self.scaling.cons_weights

        x_orig = self._orig_x(x)

        jac_orig = self.problem.cons_jac(x_orig)

        jac = jac_orig.tocoo()

        jac_row = jac.row
        jac_col = jac.col
        jac_data = jac.data

        for k, (i, j, v) in enumerate(zip(jac_row, jac_col, jac_data)):
            jac_data[k] = np.ldexp(v, cons_weights[i] - var_weights[j])

        return jac

    def lag_hess(self, x, y):
        var_weights = self.scaling.var_weights
        cons_weights = self.scaling.cons_weights
        obj_weight = self.scaling.obj_weight

        x_orig = self._orig_x(x)
        y_orig = np.ldexp(y, cons_weights - obj_weight)

        hess_orig = self.problem.lag_hess(x_orig, y_orig)

        hess = hess_orig.tocoo()

        hess_row = hess.row
        hess_col = hess.col
        hess_data = hess.data

        for k, (i, j, v) in enumerate(zip(hess_row, hess_col, hess_data)):
            combined_weight = obj_weight - var_weights[i] - var_weights[j]
            hess_data[k] = np.ldexp(v, combined_weight)

        return hess


def create_scaling(
    problem: Problem, params: Params, x0: np.ndarray, y0: np.ndarray
) -> Optional[Scaling]:
    scaling_type = params.scaling_type

    if params.scaling is not None:
        assert scaling_type == ScalingType.Custom
        return params.scaling

    if scaling_type == ScalingType.NoScaling:
        return None
    elif scaling_type == ScalingType.Custom:
        raise ValueError("Custom scaling requires explicit scaling")

    if scaling_type == ScalingType.Nominal:
        if problem.num_cons > 0:
            cons_val = problem.cons(x0)
        else:
            cons_val = np.array([], dtype=x0.dtype)
        return Scaling.from_nominal_values(x0, cons_val)

    if problem.num_cons > 0:
        cons_jac = problem.cons_jac(x0)
    else:
        cons_jac = sparse_zero(shape=(0, problem.num_vars))

    if scaling_type == ScalingType.GradJac:
        obj_grad = problem.obj_grad(x0)
        return Scaling.from_grad_jac(obj_grad, cons_jac)
    elif scaling_type == ScalingType.KKT:
        lag_hess = problem.lag_hess(x0, y0)
        return Scaling.from_equilibrated_kkt(lag_hess, cons_jac)
    else:
        raise ValueError(f"Unknown scaling type {scaling_type}")



from enum import Enum, auto

import numpy as np


class EventResultType(Enum):
    CONVERGED = auto()
    UNBOUNDED = auto()
    FILTER_CHANGED = auto()
    FREE_GRAD_ZERO = auto()
    PENALTY = auto()


class EventResult:
    def __init__(self, t: float, z: np.ndarray):
        self.t = t
        self.z = z


class ConvergedResult(EventResult):
    def __init__(self, t: float, z: np.ndarray):
        super().__init__(t, z)
        self.type = EventResultType.CONVERGED


class UnboundedResult(EventResult):
    def __init__(self, t: float, z: np.ndarray):
        super().__init__(t, z)
        self.type = EventResultType.UNBOUNDED


class FilterChangedResult(EventResult):
    def __init__(self, t: float, z: np.ndarray, filter: np.ndarray, j: int):
        super().__init__(t, z)
        self.type = EventResultType.FILTER_CHANGED

        next_filter = np.copy(filter)
        (size,) = filter.shape
        assert 0 <= j < size
        next_filter[j] = not (filter[j])
        self.filter = next_filter


class PenaltyResult(EventResult):
    def __init__(self, t: float, z: np.ndarray):
        super().__init__(t, z)
        self.type = EventResultType.PENALTY



from typing import Optional

import numpy as np
from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus
from pyomo.contrib.pynumero.linalg.ma57_interface import MA57

from .linear_solver import LinearSolver, LinearSolverError


class MA57Solver(LinearSolver):
    def __init__(self, mat, symmetric=False, report_rcond=False):
        super().__init__(mat, symmetric=symmetric)

        self.mat = mat.tocoo()
        self.solver = MA57()
        self.report_rcond = report_rcond

        # if report_rcond:
        #     self.solver.set_icntl(10, 1)

        size = self.mat.shape[0]

        if size == 0:
            self.solver = None
            return

        def symbolic():
            return self.solver.do_symbolic_factorization(self.mat, raise_on_error=False)

        self._try_fact(symbolic)

        def numeric():
            return self.solver.do_numeric_factorization(self.mat, raise_on_error=False)

        self._try_fact(numeric)

    def _try_fact(self, func):
        num_tries = 10
        for _ in range(num_tries):
            status = func()

            if status.status == LinearSolverStatus.not_enough_memory:
                self.solver.increase_memory_allocation()
                continue

            self._handle_fact_status(status)
            break
        else:
            raise Exception("Failed to compute symbolic factorization")

    def _handle_fact_status(self, status):
        if status.status == LinearSolverStatus.successful:
            return

        if status.status == LinearSolverStatus.singular:
            raise LinearSolverError("Matrix is singular")

        raise Exception("Failed to compute factorization: %s", status)

    def solve(self, rhs, trans=False, initial_sol=None):
        from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus

        if self.solver is None:
            assert rhs.size == 0
            return np.zeros_like(rhs)

        assert not trans

        x, status = self.solver.do_back_solve(rhs)

        if status.status != LinearSolverStatus.successful:
            raise LinearSolverError("Failed to compute solution")

        return x

    def num_neg_eigvals(self):
        if self.solver is None:
            return 0
        return self.solver.get_info(24)

    def rcond(self) -> Optional[float]:
        # TODO: Find out how to get rcond from MA57
        # if self.report_rcond:
        #     return self.solver.get_info(27)

        return None



from typing import List, Optional

import numpy as np

from pygradflow.callbacks import Callbacks, CallbackType
from pygradflow.display import Format, StateData, print_problem_stats, solver_display
from pygradflow.eval import EvalError
from pygradflow.iterate import Iterate
from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.penalty import penalty_strategy
from pygradflow.problem import Problem
from pygradflow.result import SolverResult
from pygradflow.status import SolverStatus
from pygradflow.step.step_control import (
    StepController,
    StepControlResult,
    step_controller,
)
from pygradflow.timer import Timer
from pygradflow.transform import Transformation

header_interval = 25


class Solver:
    """
    Solves a :py:class:`pygradflow.problem.Problem` instance according
    to the given :py:class:`pygradflow.params.Params`.
    The solver attempts to find a solution given by
    vectors :math:`x \\in \\mathbb{R}^{n}`,
    :math:`y \\in \\mathbb{R}^{m}`, and :math:`d \\in \\mathbb{R}^{n}`,
    approximately satisfying the stationarity
     .. math::
        \\begin{align}
            \\nabla_x f(x) + y^{T} J_c(x) + d = 0,
        \\end{align}

    feasibility

     .. math::
        \\begin{align}
            \\quad & l \\leq c(x) \\leq u \\\\
            \\quad & l^x \\leq x \\leq u^x \\\\
        \\end{align}

    and complementarity

     .. math::
        \\begin{align}
            y_i
            \\begin{cases}
                \\geq 0 & \\text{ if } c(x)_i = u_i \\\\
                \\leq 0 & \\text{ if } c(x)_i = l_i \\\\
                = 0 & \\text{ otherwise }
            \\end{cases} \\\\
            d_j
            \\begin{cases}
                \\geq 0 & \\text{ if } x_j = u^x_j \\\\
                \\leq 0 & \\text{ if } x_j = l^x_j \\\\
                = 0 & \\text{ otherwise }
            \\end{cases}
        \\end{align}

    conditions.
    """

    def __init__(
        self,
        problem: Problem,
        params: Params = Params(),
    ) -> None:
        """
        Creates a new solver

        Parameters
        ----------
        problem: pygradflow.problem.Problem
            The problem to be solved
        params: pygradflow.params.Params
            Parameters used by the solver
        """
        self.orig_problem = problem
        self.params = params
        self.callbacks = Callbacks()

    def compute_step(
        self,
        controller: StepController,
        iterate: Iterate,
        dt: float,
        display: bool,
        timer: Timer,
    ) -> StepControlResult:

        assert self.rho != -1.0
        return controller.compute_step(iterate, self.rho, dt, display, timer)

    def _deriv_check(self, x: np.ndarray, y: np.ndarray) -> None:
        from pygradflow.deriv_check import deriv_check
        from pygradflow.params import DerivCheck

        eval = self.evaluator
        params = self.params
        deriv_check_type = params.deriv_check

        if deriv_check_type == DerivCheck.NoCheck:
            return

        if deriv_check_type & DerivCheck.CheckFirst:
            logger.info("Checking objective derivative")
            deriv_check(lambda x: eval.obj(x), x, eval.obj_grad(x), params)

            logger.info("Checking constraint derivative")
            deriv_check(lambda x: eval.cons(x), x, eval.cons_jac(x), params)

        if deriv_check_type & DerivCheck.CheckSecond:
            logger.info("Checking Hessian")

            deriv_check(
                lambda x: eval.obj_grad(x) + eval.cons_jac(x).T.dot(y),
                x,
                eval.lag_hess(x, y),
                params,
            )

        logger.info("Finished derivative check")

    def print_result(
        self,
        total_time: float,
        status: SolverStatus,
        iterate: Iterate,
        iterations: int,
        accepted_steps: int,
        dist_factor: float,
    ) -> None:
        rho = self.rho

        desc = "{:>45s}".format(SolverStatus.description(status))

        status_desc = Format.redgreen(desc, SolverStatus.success(status), bold=True)
        status_name = Format.bold("{:>20s}".format("Status"))

        logger.info("%20s: %45s", status_name, status_desc)
        logger.info("%20s: %45s", "Time", f"{total_time:.2f}s")
        logger.info("%20s: %45d", "Iterations", iterations)
        logger.info("%20s: %45d", "Accepted steps", accepted_steps)

        logger.info("%20s: %45e", "Distance factor", dist_factor)

        logger.info("%20s: %45e", "Objective", iterate.obj)
        logger.info("%20s: %45e", "Aug Lag violation", iterate.aug_lag_violation(rho))
        logger.info("%20s: %45e", "Aug Lag dual", iterate.aug_lag_dual())

        logger.info("%20s: %45e", "Constraint violation", iterate.cons_violation)
        logger.info("%20s: %45e", "Dual violation", iterate.stat_res)

        eval = self.evaluator

        eval_name = Format.bold("{:>20s}".format("Evaluations"))
        logger.info("%20s", eval_name)

        for component, num_evals in eval.num_evals.items():
            name = component.name()
            logger.info("%20s: %45d", name, num_evals)

    def _check_terminate(self, iterate, iteration, timer):
        params = self.params

        if (params.iteration_limit is not None) and (
            iteration >= params.iteration_limit
        ):
            logger.debug("Iteration limit reached")
            return SolverStatus.IterationLimit

        if timer.reached_time_limit():
            logger.debug("Reached time limit")
            return SolverStatus.TimeLimit

        if iterate.total_res <= params.opt_tol:
            logger.debug("Convergence achieved")
            return SolverStatus.Optimal

        if iterate.locally_infeasible(params.opt_tol, params.local_infeas_tol):
            logger.debug("Local infeasibility detected")
            return SolverStatus.LocallyInfeasible

        if (iterate.obj <= params.obj_lower_limit) and (
            iterate.is_feasible(params.opt_tol)
        ):
            logger.debug("Unboundedness detected")
            return SolverStatus.Unbounded

    def solve(
        self,
        x0: np.ndarray | float | None = None,
        y0: np.ndarray | float | None = None,
    ) -> SolverResult:
        """
        Solves the problem starting from the given primal / dual point

        Parameters
        ----------
        x0: np.ndarray | float | None
            The initial primal point :math:`x_0 \\in \\mathbb{R}^{n}`
        y0: np.ndarray | float | None
            The initial dual point :math:`y_0 \\in \\mathbb{R}^{m}`

        Returns
        -------
        pygradflow.solver.SolverResult
            The result of the solving process, including primal and dual
            solutions
        """

        self.transform = Transformation(self.orig_problem, self.params, x0, y0)

        self.problem = self.transform.trans_problem

        params = self.params
        problem = self.problem

        self.evaluator = self.transform.evaluator

        self.penalty = penalty_strategy(self.problem, params)
        self.rho = -1.0

        display = solver_display(problem, params)

        iterate = self.transform.initial_iterate

        try:
            iterate.check_eval()
        except EvalError as e:
            raise Exception("Failed to evaluate initial iterate") from e

        print_problem_stats(problem, iterate)

        lamb = params.lamb_init

        controller = step_controller(problem, params)

        self._deriv_check(iterate.x, iterate.y)

        self.rho = self.penalty.initial(iterate)

        logger.debug("Initial Aug Lag: %.10e", iterate.aug_lag(self.rho))

        status = None
        iteration = 0

        logger.info(display.header)

        path_dist = 0.0
        initial_iterate = iterate
        accepted_steps = 0
        iteration = 0

        timer = Timer(params.time_limit)

        if params.collect_path:
            path: Optional[List[np.ndarray]] = [initial_iterate.z]
            path_times = [0.0]
        else:
            path = None

        while True:
            status = self._check_terminate(iterate, iteration, timer)
            if status is not None:
                break

            display_iterate = display.should_display()

            step_result = self.compute_step(
                controller, iterate, 1.0 / lamb, display_iterate, timer
            )

            x = iterate.x
            y = iterate.y

            next_iterate = step_result.iterate
            accept = step_result.accepted
            lamb = step_result.lamb

            if lamb >= params.lamb_max:
                raise Exception(
                    f"Inverse step size {lamb} exceeded maximum {params.lamb_max} (incorrect derivatives?)"
                )

            primal_step_norm = float(np.linalg.norm(next_iterate.x - iterate.x))
            dual_step_norm = float(np.linalg.norm(next_iterate.y - iterate.y))

            self.callbacks(CallbackType.ComputedStep, iterate, next_iterate, accept)

            if display_iterate:
                state = StateData()
                state["iterate"] = iterate
                state["active_set"] = lambda: step_result.active_set
                state["obj_nonlin"] = lambda: iterate.obj_nonlin(next_iterate)

                if problem.num_cons > 0:
                    state["cons_nonlin"] = lambda: np.linalg.norm(
                        iterate.cons_nonlin(next_iterate), ord=np.inf
                    )

                state["aug_lag"] = lambda: iterate.aug_lag(self.rho)
                state["obj"] = lambda: iterate.obj()
                state["iter"] = iteration + 1
                state["primal_step_norm"] = lambda: primal_step_norm
                state["dual_step_norm"] = lambda: dual_step_norm
                state["lamb"] = lambda: lamb
                state["step_accept"] = lambda: accept
                state["rcond"] = lambda: step_result.rcond

                logger.info(display.row(state))

            if accept:
                penalty_result = self.penalty.update(iterate, next_iterate)
                next_rho = penalty_result.next_rho
                accept = penalty_result.accept

            if accept:

                if next_rho != self.rho:
                    logger.debug(
                        "Updating penalty parameter from %e to %e", self.rho, next_rho
                    )
                    self.rho = next_rho

                if path is not None:
                    path.append(next_iterate.z)
                    path_times.append(path_times[-1] + (1.0 / lamb))

                iterate = next_iterate

                path_dist += primal_step_norm + dual_step_norm
                accepted_steps += 1

            iteration += 1

        total_time = timer.elapsed()

        direct_dist = iterate.dist(initial_iterate)

        assert (path_dist >= direct_dist) or (np.isclose(path_dist, direct_dist))

        dist_factor = path_dist / direct_dist if direct_dist != 0.0 else 1.0
        iterations = iteration

        assert status is not None

        self.print_result(
            total_time=total_time,
            status=status,
            iterate=iterate,
            iterations=iterations,
            accepted_steps=accepted_steps,
            dist_factor=dist_factor,
        )

        x = iterate.x
        y = iterate.y
        d = iterate.bounds_dual

        (x, y, d) = self.transform.restore_sol(x, y, d)

        result = SolverResult(
            problem,
            x,
            y,
            d,
            status,
            iterations=iterations,
            num_accepted_steps=accepted_steps,
            total_time=total_time,
            dist_factor=dist_factor,
            final_scaled_obj=iterate.obj,
            final_stat_res=iterate.stat_res,
            final_cons_violation=iterate.cons_violation,
        )

        if path is not None:
            complete_path = np.vstack(path).T
            model_times = np.hstack(path_times)
            result._set_path(complete_path, model_times)

        return result



import functools
from typing import Tuple

import numpy as np

from pygradflow.cons_problem import ConstrainedProblem
from pygradflow.eval import create_evaluator
from pygradflow.iterate import Iterate
from pygradflow.log import logger
from pygradflow.params import Params
from pygradflow.problem import Problem
from pygradflow.scale import ScaledProblem, create_scaling


class Transformation:
    def __init__(
        self,
        orig_problem: Problem,
        params: Params,
        x0: np.ndarray | float | None,
        y0: np.ndarray | float | None,
    ):
        self.orig_problem = orig_problem

        (self.x0, self.y0) = self._create_initial_values(orig_problem, x0, y0)
        self.scaling = create_scaling(orig_problem, params, self.x0, self.y0)
        self.params = params

        self.evaluator = create_evaluator(self.trans_problem, params)

    def _create_initial_values(
        self,
        orig_problem,
        x0: np.ndarray | float | None,
        y0: np.ndarray | float | None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        orig_lb = orig_problem.var_lb
        orig_ub = orig_problem.var_ub

        if x0 is None:
            x0 = np.clip(0.0, orig_problem.var_lb, orig_problem.var_ub)
        else:
            if (x0 > orig_ub).any() or (x0 < orig_lb).any():
                logger.warning("Initial point violates variable bounds")
                x0 = np.clip(x0, orig_lb, orig_ub)

        if y0 is None:
            y0 = 0.0

        x0 = np.broadcast_to(x0, shape=(orig_problem.num_vars,))
        y0 = np.broadcast_to(y0, shape=(orig_problem.num_cons,))

        return (x0, y0)

    @functools.cached_property
    def scaled_problem(self):
        orig_problem = self.orig_problem
        scaling = self.scaling

        if scaling is None:
            return orig_problem

        return ScaledProblem(orig_problem, scaling)

    @functools.cached_property
    def trans_problem(self):
        scaled_problem = self.scaled_problem

        return ConstrainedProblem(scaled_problem)

    @property
    def initial_iterate(self):
        (x0, y0) = self.transform_sol(self.x0, self.y0)
        dtype = self.params.dtype

        x = x0.astype(dtype)
        y = y0.astype(dtype)

        return Iterate(self.trans_problem, self.params, x, y, self.evaluator)

    def transform_sol(self, x: np.ndarray, y: np.ndarray):
        scaling = self.scaling

        if scaling is None:
            scaled_x = x
            scaled_y = y
        else:
            scaled_x = scaling.scale_primal(x)
            scaled_y = scaling.scale_dual(y)

        trans_problem = self.trans_problem

        return trans_problem.transform_sol(scaled_x, scaled_y)

    def restore_sol(self, x: np.ndarray, y: np.ndarray, d: np.ndarray):
        trans_problem = self.trans_problem

        (x, y, d) = trans_problem.restore_sol(x, y, d)

        scaling = self.scaling

        if scaling is None:
            return (x, y, d)

        orig_x = scaling.unscale_primal(x)
        orig_y = scaling.unscale_dual(y)
        orig_d = scaling.unscale_bounds_dual(d)

        return (orig_x, orig_y, orig_d)



import abc
from typing import Optional

import numpy as np
import scipy as sp

from pygradflow.iterate import Iterate
from pygradflow.problem import Problem
from pygradflow.util import keep_rows


class StepFunc(abc.ABC):
    def __init__(self, problem: Problem, iterate: Iterate, dt: float) -> None:
        self.problem = problem
        self.orig_iterate = iterate
        self.dt = dt

        self.n = problem.num_vars
        self.m = problem.num_cons

    def compute_active_set_box(self, x: np.ndarray, lb: np.ndarray, ub: np.ndarray):
        """
        Compute the active set at the given point with respect to the
        given bounds.

        Parameters
        ----------

        x: np.ndarray
            A primal point in :math:`\\mathbb{R}^{n}`
        lb, ub: np.ndarray
            Bounds in :math:`\\mathbb{R}^{n}`

        Returns
        -------
        np.ndarray
            A vector with boolean entries, where a value of `True`
            at index `j` indicates that :math:`x_j` is
            outside the interval  :math:`[l_j, u_j]` and should be
            clipped to it during the projection.
        """
        assert (lb <= ub).all()

        return np.logical_or(x < lb - 1e-8, x > ub + 1e-8)

    def project_box(
        self, x: np.ndarray, lb: np.ndarray, ub: np.ndarray, active_set: np.ndarray
    ) -> np.ndarray:
        assert active_set.dtype == bool
        assert active_set.shape == lb.shape
        assert (lb <= ub).all()

        p = np.copy(x)

        p[active_set] = np.clip(x[active_set], lb[active_set], ub[active_set])

        assert (lb[active_set] <= p[active_set]).all()
        assert (p[active_set] <= ub[active_set]).all()

        return p

    @abc.abstractmethod
    def value_at(
        self, iterate: Iterate, rho, active_set: Optional[np.ndarray] = None
    ) -> np.ndarray:
        raise NotImplementedError()

    @abc.abstractmethod
    def projection_initial(self, iterate: Iterate, rho: float, tau=None) -> np.ndarray:
        raise NotImplementedError()

    def compute_active_set(self, iterate: Iterate, rho: float, tau=None) -> np.ndarray:
        p = self.projection_initial(iterate, rho, tau)
        return self.active_set_at_point(p)

    @abc.abstractmethod
    def active_set_at_point(self, p: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def apply_project_deriv(self, mat: sp.sparse.spmatrix, active_set: np.ndarray):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub
        (num_rows, _) = mat.shape

        assert active_set.dtype == bool
        assert active_set.shape == lb.shape

        assert (lb <= ub).all()

        assert (lb != np.inf).all()
        assert (ub != -np.inf).all()

        assert (num_rows,) == lb.shape

        inactive_set = np.logical_not(active_set)
        proj_mat = keep_rows(mat, inactive_set)

        return proj_mat


class ImplicitFunc(StepFunc):
    def __init__(self, problem: Problem, iterate: Iterate, dt: float) -> None:
        super().__init__(problem, iterate, dt)

    def project(self, x: np.ndarray, active_set: np.ndarray):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub

        return super().project_box(x, lb, ub, active_set)

    def active_set_at_point(self, p):
        problem = self.problem
        lb = problem.var_lb
        ub = problem.var_ub
        return super().compute_active_set_box(p, lb, ub)

    def projection_initial(self, iterate: Iterate, rho: float, tau=None):
        x_0 = self.orig_iterate.x
        dt = self.dt

        if tau is not None:
            lamb = 1.0 / dt
            x = iterate.x
            return (
                (1.0 - tau * lamb) * x
                + (tau * lamb) * x_0
                - tau * iterate.aug_lag_deriv_x(rho)
            )

        return x_0 - dt * iterate.aug_lag_deriv_x(rho)

    # @override
    def value_at(self, iterate, rho, active_set=None):
        y_0 = self.orig_iterate.y
        dt = self.dt

        p = self.projection_initial(iterate, rho)

        if active_set is None:
            active_set = self.compute_active_set(iterate, rho)

        xval = iterate.x - self.project(p, active_set)
        yval = iterate.y - (y_0 + dt * iterate.aug_lag_deriv_y())
        return np.concatenate([xval, yval])

    def deriv(
        self, jac: sp.sparse.spmatrix, hess: sp.sparse.spmatrix, active_set: np.ndarray
    ):
        n = self.n
        m = self.m
        dt = self.dt

        assert active_set is not None
        params = self.orig_iterate.params

        F_11 = sp.sparse.eye(n, dtype=params.dtype)
        F_11 += self.apply_project_deriv(dt * hess, active_set)

        F_12 = self.apply_project_deriv(dt * jac.T, active_set)

        assert F_11.dtype == params.dtype
        assert F_12.dtype == params.dtype

        F_21 = -dt * jac
        F_22 = sp.sparse.eye(m, dtype=params.dtype)

        deriv = sp.sparse.bmat([[F_11, F_12], [F_21, F_22]], format="csc")

        assert deriv.dtype == params.dtype

        return deriv

    def deriv_at(
        self, iterate: Iterate, rho: float, active_set: Optional[np.ndarray] = None
    ):
        if active_set is None:
            active_set = self.compute_active_set(iterate, rho)

        hess = iterate.aug_lag_deriv_xx(rho)
        jac = iterate.aug_lag_deriv_xy()

        return self.deriv(jac, hess, active_set)


class ScaledImplicitFunc(StepFunc):
    def __init__(self, problem: Problem, iterate: Iterate, dt: float) -> None:
        super().__init__(problem, iterate, dt)
        self.lamb = 1.0 / dt
        params = iterate.params
        self.lb = (self.lamb * problem.var_lb).astype(params.dtype)
        self.ub = (self.lamb * problem.var_ub).astype(params.dtype)

    # @override
    def value_at(self, iterate, rho, active_set=None):
        y_0 = self.orig_iterate.y
        lamb = self.lamb

        p = self.projection_initial(iterate, rho)

        if active_set is None:
            active_set = self.compute_active_set(iterate, rho)

        xval = lamb * iterate.x - self.project(p, active_set)
        yval = -(lamb * iterate.y - (lamb * y_0 + iterate.aug_lag_deriv_y()))

        return np.concatenate([xval, yval])

    def projection_initial(self, iterate: Iterate, rho: float, tau=None):
        x_0 = self.orig_iterate.x
        lamb = self.lamb

        if tau is not None:
            lamb = 1.0 / self.dt
            x = iterate.x
            f_x = lamb * (1 - tau * lamb)
            f_x0 = tau * lamb * lamb
            f_d = tau * lamb

            return f_x * x + f_x0 * x_0 - f_d * iterate.aug_lag_deriv_x(rho)

        return lamb * x_0 - iterate.aug_lag_deriv_x(rho)

    def project(self, x: np.ndarray, active_set: np.ndarray):
        return super().project_box(x, self.lb, self.ub, active_set)

    def active_set_at_point(self, p):
        return super().compute_active_set_box(p, self.lb, self.ub)

    def deriv(
        self, jac: sp.sparse.spmatrix, hess: sp.sparse.spmatrix, active_set: np.ndarray
    ):
        n = self.n
        m = self.m
        dt = self.dt
        lamb = 1.0 / dt

        assert active_set is not None
        params = self.orig_iterate.params

        F_11 = sp.sparse.diags([lamb], shape=(n, n), dtype=params.dtype)

        F_11 = lamb * sp.sparse.eye(n, dtype=params.dtype)
        F_11 += self.apply_project_deriv(hess, active_set)

        F_12 = self.apply_project_deriv(jac.T, active_set)

        assert F_11.dtype == params.dtype
        assert F_12.dtype == params.dtype

        F_21 = -jac

        F_22 = sp.sparse.diags([lamb], shape=(m, m), dtype=params.dtype)

        deriv = sp.sparse.bmat([[F_11, F_12], [F_21, F_22]], format="csc")

        assert deriv.dtype == params.dtype

        return deriv

    def deriv_at(
        self, iterate: Iterate, rho: float, active_set: Optional[np.ndarray] = None
    ):
        if active_set is None:
            active_set = self.compute_active_set(iterate, rho)

        hess = iterate.aug_lag_deriv_xx(rho)
        jac = iterate.aug_lag_deriv_xy()

        return self.deriv(jac, hess, active_set)



import sksparse.cholmod

from .linear_solver import LinearSolver, LinearSolverError


class CholeskySolver(LinearSolver):
    def __init__(self, mat, symmetric=False):
        super().__init__(mat, symmetric=symmetric)

        try:
            self.factor = sksparse.cholmod.cholesky(mat)
            self.factor.L()
        except sksparse.cholmod.CholmodNotPositiveDefiniteError as e:
            raise LinearSolverError() from e

    def solve(self, rhs, trans=False, initial_sol=None):
        assert not trans

        return self.factor.solve_A(rhs)

    def num_neg_eigvals(self):
        return 0



import scipy as sp

from .linear_solver import LinearSolver, LinearSolverError


class MINRESSolver(LinearSolver):
    def __init__(self, mat, symmetric=False):
        super().__init__(mat, symmetric=symmetric)
        assert symmetric, "MINRES requires a symmetric matrix"
        self.mat = mat

    def solve(self, rhs, trans=False, initial_sol=None):
        # matrix should be symmetric anyways
        if initial_sol is not None:
            initial_sol = initial_sol()

        result = sp.sparse.linalg.minres(self.mat, rhs, x0=initial_sol)

        (sol, info) = result

        if info != 0:
            raise LinearSolverError("MINRES failed with error code {}".format(info))

        return sol
